/***************************************************************************
 *   QGame++                                                               *
 *   =======                                                               *
 *   A Quantum Gate And Measurement Emulator.                              *
 *                                                                         *
 *   Copyright (C) 2003/04 by Manuel Nickschas                             *
 *   <qgame-devel@nickschas.de>                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "qgame++.h"

using namespace std;
using namespace qgame;

ostream & qgame::operator<<(ostream &o, const Result &r) {
  o << "Misses: " << r.misses << endl;
  o << "Max error: " << r.maxError << endl;
  o << "Average error: " << r.avgError << endl;
  o << "Max expected oracles: " << r.maxExpOracles << endl;
  o << "Average expected oracles: " << r.avgExpOracles << endl;
  return o;
}
  

QSys::QSys() {
  prg = NULL;
  numQubits = 0;
  oracleTT = "";
}

QSys::~QSys() {
  delete prg;
}

void QSys::init (int numqb, QProgram *p, const string &tt) {
  prg = p;
  instrPnt = 0;
  oracleTT = tt;
  prob = 1;
  numQubits = numqb;
  reg = Qureg(numqb);
  prg->installOracle(tt);
  oracleCount = 0;
  execInstr.clear();
  measureHistory.clear();
  execInstr = vector<bool>(p->length(), true);
};

void QSys::setReg(const Qureg &r) {
  reg = r;
}

Qureg QSys::getReg() {
  return reg;
}

double QSys::getProb() {
  return prob;
}

void QSys::setProb(double p) {
  prob = p;
}

int QSys::getOracleCount() {
  return oracleCount;
}

void QSys::appendMeasurement(QubitList q, int outcome, double prob) {
  Measurement m;
  m.qubit = q[0];
  m.result = outcome;
  m.prob = prob;
  measureHistory.push_back(m);
}

QSys *QSys::clone(double p) {
  QSys *res = new QSys;
  res->prob      = p;
  res->instrPnt  = instrPnt;
  res->prg       = prg;
  res->oracleTT  = oracleTT;
  res->numQubits = numQubits;
  res->reg       = reg;
  res->oracleCount = oracleCount;
  res->execInstr = execInstr;
  res->measureHistory = measureHistory;
  return res;
}

void QSys::apply(QGate *gate, const QubitList &qubits) throw(Error) {
  if(gate->numQb() != qubits.size()) throw(Error(E_MISMATCH));
  vector<QuSubReg> subregs = reg.getSubRegs(qubits);
  for(int sr = 0; sr < subregs.size(); sr++) {
    gate->apply(subregs[sr]);
  }
}

void QSys::apply(const QInstr &instr) throw(Error) {
  if(!instr.gate) return;
  if(instr.type == ORACLE) oracleCount++;
  else if(instr.type == LIMITED_ORACLE) {
    if(oracleCount >= instr.extra) return;
    else oracleCount++;
  }
  apply(instr.gate, instr.qubits);
}

vector<QSys*> QSys::execute(int numqb, QProgram *p, const string &tt) throw(Error) {
  QSys *sys = new QSys;
  sys->init(numqb, p, tt);
  return sys->exec();
}

vector<QSys*> QSys::exec() throw(Error) {
  vector<QSys*> result;
  bool term = false;
  while(!term && instrPnt<prg->length()) {
    if(!execInstr[instrPnt]) { instrPnt++; continue; }
    QInstr *instr = (*prg)[instrPnt];
    if(instr->gate) {
      //cerr << "\napplying: "<< *instr;
      apply(*instr);
      instrPnt++;
    } else {
      instrPnt++;
      switch(instr->type) {
        case MEASURE: {
          vector<double> probs = reg.getProbabilities(instr->qubits);
          QSys *ifsys = NULL;
          QSys *elsys = NULL;
          if(probs[0]>0) elsys = this;
          if(probs[1]>0) {
            if(elsys) ifsys = clone(prob);
            else ifsys = this;
          }
          if(ifsys) { 
            ifsys->reg.collapse(instr->qubits, 1);
            ifsys->setProb(ifsys->getProb() * probs[1]);
            ifsys->appendMeasurement(instr->qubits, 1, probs[1]);
          }
          if(elsys) {
            elsys->reg.collapse(instr->qubits, 0);
            elsys->setProb(elsys->getProb() * probs[0]);
            elsys->appendMeasurement(instr->qubits, 0, probs[0]);
          }
          /* mark the appropriate branches as invalid by setting the execInstr[] flag to false */
          int i = instrPnt;
          int d = 0; 
          bool flg = false; bool branch = true;
          QSys *curSys = elsys;
          while(!flg && i<prg->length()) {
            if(curSys) curSys->execInstr[i] = false;
            switch((*prg)[i]->type) {
            case MEASURE:   // handle nested measurements
              d+=2; break;
            case END:
              if(d>0) d--;  // belongs to a nested measurement -> ignore
              else {
                if(branch) {
                  if(elsys) elsys->instrPnt = i+1;
                  if(!ifsys) flg = true;
                  else { branch = false; curSys = ifsys; }
                } else flg = true;  // we are done with marking...
              }
              break;
            }
            i++;
          }
          if(ifsys && elsys) {
            vector<QSys*> ifRes = ifsys->exec();  // execute if-branch
            for(int r = 0; r<ifRes.size(); r++) result.push_back(ifRes[r]);
          }
          break;
        }
        case HALT:
          term = true;
          break;
          
        case ORACLE:
          throw(Error(E_NOORACLE));
          break;
        
        case PRINTAMPS:
          //cout << "\nQuReg contents after measuring";
          //for(int i = 0; i<measureHistory.size(); i++) {
          //  cout << "\n  q" <<measureHistory[i].qubit << " = " << measureHistory[i].result
          //       << " (p = " << measureHistory[i].prob << ")";
          //}
          //cout << "\n" << reg;
          cout << *this;
          break;
        //default:
          
      }   
    }
  }
  result.push_back(this);
  return result;
}

Result QSys::testProgram(int numqb, QProgram *p, const std::vector<TestCase> &cases, 
                    const QubitList &finm, double thresh) throw(Error) {
  Result res;
  res.misses = 0;
  res.avgError = res.avgExpOracles = res.maxError = res.maxExpOracles = 0;
  double totalErr = 0;
  double totalExpOracles = 0;
  for(int c = 0; c < cases.size(); c++) {
    TestCase tcase = cases[c];
    vector<QSys*> systems = execute(numqb, p, tcase.oracleTT);
    double expOracles = 0;
    double rawerr = 1.0;
    for(int s=0; s<systems.size(); s++) {
      rawerr -= systems[s]->getReg().getProbabilities(finm)[tcase.desiredRes] * systems[s]->getProb();
      expOracles += systems[s]->getProb() * systems[s]->getOracleCount();  
    }
    if(rawerr<-0.000000001) { cerr << "\nWARNING: rawerr = "<<rawerr; } // DEBUG
    if(rawerr>thresh) res.misses++;
    totalErr += rawerr;
    if(rawerr > res.maxError) res.maxError = rawerr;
    totalExpOracles += expOracles;
    if(expOracles > res.maxExpOracles) res.maxExpOracles = expOracles;
  }
  res.avgError = totalErr / cases.size();
  res.avgExpOracles = totalExpOracles / cases.size();
  return res;
}

ostream& qgame::operator<<(std::ostream &o, const QSys &s) {
  o << "\n========================================================================";
  o << "\nQSys ("<<s.numQubits<<" qubits, prior prob = "<<s.prob<<")\n";
  o << "\nMeasurement History:";
  
  for(int i = 0; i<s.measureHistory.size(); i++) {
    o << "\n   q" <<s.measureHistory[i].qubit << " = " << s.measureHistory[i].result 
         << " (p = " << s.measureHistory[i].prob << ")";
  }
  o << "\nRegister:" << s.reg << "\n";
  
  return o;

}

/*****************************************************************************
 * Documentation - To be used by Doxygen. Please see the autogenerated
 * docs in the doc/ subdir for a formatted version of this ;-)
 *****************************************************************************/

/**! \mainpage The QGame++ System

\section intro Introduction
   
QGame++ (Quantum Gate And Measurement Emulator) is a system that allows
a user to run quantum computing algorithms
on an ordinary digital computer. Because quantum computers have complexity
advantages over classical computers, any classical emulator will necessarily
be less efficient than the quantum computer that it is emulating. 
QGame++ nonetheless allows the user to find out what outputs the quantum
program would produce, and with what probabilities (since quantum
computation is in general not deterministic). 

QGame++ is based on QGAME, which was developed in Common Lisp by 
Lee Spector (lspector@hampshire.edu). The QGAME homepage can be found
at http://hampshire.edu/lspector/qgame.html . For performance reasons,
a C++ version of QGAME seemed to be useful, which is why QGame++ was
developed.
Both QGAME and QGame++ were developed for use in conjunction with a
genetic programming system, to allow for the evolution of new quantum
algorithms, but can be useful for testing quantum algorithms regardless 
of their origin.

QGame++ is based on the "quantum gate array" model of quantum computation,
in which quantum "gates" (represented as square matrices) are applied
to a register of qubits (via tensor product formation and matrix
multiplication). QGame++ always starts with all qubits having the
value zero (in the state |00...0>), applies a sequence of gates,
and returns values about the resulting state. Measurement gates
cause the system to branch, following one execution path (with the
appropriate quantum state collapse) for each possible value. Final
measurements are made across the end-states of all of the resulting
branches.

Most interesting quantum algorithms involve calling an "oracle" or
"black box" function of which one is trying to determine some
property. QGame++ supports only boolean oracles (that is, with single
qubit output), but it allows the user to specify any such oracle
and to indicate the desired system output for each specified
oracle. Oracles are implemented in the standard way, by applying
a (unitary) quantum "NOT" gate on the output qubit wherever the
oracle's truth table indicates a "1".

More extensive documentation concerning quantum programs and the
quantum gate array model, as well as general information concerning
the QGAME system and its uses in genetic programming can be found in
Lee Spector's soon-to-appear book:

Spector, Lee. (2004). <em>Automatic Quantum Computer Programming: A Genetic 
Programming Approach.</em> Kluwer Academic Publishers.

\section libqgame The QGame++ Library

The core of the QGame++ system is libqgame++, a shared library that
can be used from other programs. The functions and data structures it provides
can be made available by including the file qgame++.h. This documentation
should be sufficient to show you how to use libqgame++ in your own 
programs.

\section qgamecli The qgame Command-Line Client

qgame is a simple command-line client that allows you to access 
the functionality provided by libqgame++ from the command line.
It reads a QGAME program from the command line or STDIN and runs
it either in evaluation mode (where the program is run several times
on different test cases and in the end some statistical data is produced)
or simply executes it once (outputting the list of resulting quantum
systems in the end, including their respective probabilities and 
register contents). The evaluation mode effectively runs QSys::testProgram()
provided by libqgame++, and is equivalent to the TEST-QUANTUM-PROGRAM function
in the Lisp version of the QGAME system.

qgame parses text files containing QGAME programs
in the syntax defined by Lee Spector as described in the \ref qgame_lang_desc,
or in the book on automatic quantum programming cited above.
Use 'qgame --help' or 'man qgame' for acceptable command line options. 

You might also want to look in the qgame code to see how libqgame++ is 
used, which should help you getting started with your own libqgame++  
project. Note that the focus of this project has been to provide a 
robust libqgame++ library that can be used in client applications, not 
to provide a polished command-line interface. The qgame command-line 
interface provided here serves as a simple demonstration of the use of 
the libqgame++ library, and may also be useful as-is, but it is fairly 
simple and has not been extensively documented.

\section install Requirements and Installation

QGame++ should run on any system supporting the current C++ standard.
It has been developed on a Linux system using current versions of gcc 3.x,
automake and related tools, and most testing was done on Linux, too.
However, there have been successes using MacOS X, and it probably will
even run on Windows (although I can't help you with compilation and
installation there).

Installation instructions can be found in INSTALL. If you are familiar
with the standard autoconf/automake style of configuration and installation,
you probably won't need to look there at all and can just go ahead using
'./configure', 'make' and 'make install' as usual. This will install libqgame++,
its header files, the command line client and a man page for qgame in the
standard locations on your system.

\section disclaimer Some Important Remarks

QGame++ is still work in progress, and in an early beta stage. Although it
is believed to work correctly, it has not been tested extensively so far.
Also I expect the interfaces and QGAME syntax to change in later versions,
because the integration in Lee Spector's genetic programming system PushGP
(http://hampshire.edu/lspector/push.html) will quite possibly lead to 
modifications of the current version. So, even though QGame++ is already 
quite usable in its current state, be warned that there is no guarantee
that later versions might break compatibility.

If you find any bugs or problems in this program, or if you want to discuss
new features or modifications to existing features, feel free to send a 
message to qgame-devel@nickschas.de. I would also like to hear from you if
you successfully use QGame++ in a project. In general, any comment is welcomed.

The QGame++ system is licensed under the GPL, and you can find the licensing
conditions implied by the GPL in the file COPYING supplied with the source code.
There is <em>no</em> warranty; not even for <em>merchantibility</em> or
<em>fitness for a particular purpose</em>.

\section acknowledgements Acknowledgements

As mentioned above, QGame++ is based on Lee Spector's Lisp version QGAME,
and even though naturally the C++ code is very different from the Lisp code
and I used a different approach to implement most things, this work was heavily
inspired by QGAME. In addition, planned as a faster replacement for QGAME, the
syntax used and features that have been implemented come straight from QGAME,
as do parts of this documentation. Thanks and credits therefore go to Lee Spector,
whose homepage at http://hampshire.edu/lspector you should not miss.

This work was also funded by Lee, who therefore helped me survive in a foreign
country ;-), and gave me the opportunity to be involved in research in the U.S.,
in spite of the numerous difficulties arising from my status as an exchange
student. Many thanks for that.

-- Manuel Nickschas \<qgame-devel@nickschas.de><br>
Amherst, Massachusetts<br>
March 2004
 */
 
/**!
\page qgame_lang_desc QGAME Language Description
 
QGAME is a language used to describe quantum programs using the gate model.
The syntax that can be parsed by QGame++ follows that used by Lee Spector's 
original QGAME implementation, and is also described in his soon-to-be-published
book "Automatic Quantum Computer Programming: A Genetic Programming Approach"
and various material found on Lee's homepage at http://hampshire.edu/lspector .
I will give a brief description of the syntax here.

\note
The current version of the parser is line-oriented, meaning that each instruction must
be in a single line. The parser currently also is not very robust when it comes
to extra whitespace and so on. Note that libqgame++ does not care for how its
QProgram object is created, and you might write your own parser that spills out
QInstruction objects, or create a QProgram any way you want. This language
description therefore is used mostly to describe what to feed into the command
line client 'qgame' rather than how to access libqgame++ itself.

\section qgame_params Parameters

The parameters needed for program execution can be supplied on the command line,
or specified in the QGAME program file. For the latter, parameter lines are started
with a colon, followed by the parameter's name and a space-separated list of values.
Note that parameters given on the command line override those found in the text.

- <b>:num-qubits \<num></b> defines the number of qubits the system should use for executing
  the program.

- <b>:cases \<case_1> \<case_2> ... \<case_n></b> supplies a number of test cases used
  for program evaluation if qgame is run in evaluation mode. A test case specifies 
  the truth table and the desired result for this run. '0101-1' therefore means
  that the oracle's truth table is 0101, and one wants to measure a 1. The probability
  for measuring the desired result in each test case will define the evaluation result.
  
- <b>:threshold \<t></b> defines the error threshold to be used when evaluating the
  program. If the error probability of measuring the desired result is below that
  threshold, the test run is considered to be successful. A common value used here
  is 0.48.
  
- <b>:final-measurement-qubits \<q1> \<q2> ... \<qn></b> specifies the list of
  qubits that are to be measured after execution of the program in order to achieve
  the desired result given for the current test case. The least significant qubit is listed
  last. The resulting probability is computed across all possible branches caused by
  intermediate measurements.
  
\section qgame_syntax QGAME Instructions

After specifying parameters, a list of QGAME instructions should follow, each instruction
in a single line (no linebreaks). In addition, you might have comment lines started by a '#'.
Note that QGame++ is based on a LISP program and the syntax therefore looks pretty LISPish
(i.e. there are a lot of brackets...).

\subsection qgame_syntax_gates Quantum Gate Instructions

Following is a list of all QGAME instructions that define gates. Note that whenever
a qubit list is specified, a list of integers is expected, with the least significant
qubit last.

- <b>(QNOT \<q>)</b> applies a quantum NOT gate to the specified qubit

- <b>(CNOT \<control> \<target>)</b> applies a quantum controlled NOT gate to the
  specified control and target qubits

- <b>(SRN \<q>)</b> applies a quantum square-root-of-NOT gate to the specified qubit

- <b>(NAND \<in1> \<in2> \<out>)</b> applies a quantum NAND gate to the specified
  input and output qubits

- <b>(HADAMARD \<q>)</b> applies a Hadamard gate to the specified qubit

- <b>(U-THETA \<q> \<theta>)</b> applies a rotation gate with the specified (real-valued)
  angle theta to the specified qubit

- <b>(CPHASE-OLD \<control> \<target> \<alpha>)</b>: one version of a controlled phase
  gate -- see \ref qgates for the matrix

- <b>(CPHASE \<control> \<target> \<alpha>)</b>: another (probably preferable)
  version of a controlled phase gate -- see \ref qgates for the matrix
     
- <b>(U2 \<q> \<phi> \<theta> \<psi> \<alpha>)</b>: a general rotation gate for a
  single qubit with 4 real-valued parameters -- see \ref qgates for the matrix

- <b>(SWAP \<q1> \<q2>)</b> applies a gate that swaps the amplitudes of the two specified qubits

- <b>(MATRIX-GATE #2A((row1)(row2)...(rown)) (history))</b>: This specifies a general
  matrix gate. The matrix is given as a LISP-style 2D array. Each row contains space-separated
  real values. Note that although a matrix technically can contain complex values, this
  is not supported by the current version of the parser (you might create QMatrixGate objects
  containing complex values though, if you use libqgame++ in your own program or create
  a QProgram object by other means). The specified matrix must be unitary and be of size
  2^n X 2^n, with n being the number of qubits in the system. The gate is applied to the complete
  quantum register. (history) is an optional parameter that is not currently used by QGame++.

- <b>(ORACLE ORACLE-TT \<q1> \<q2> ... \<qn> \<out>)</b>:
  calls the oracle on the specified input qubits (specified as "q"s 
  above) and the specified output qubit. The input qubits are listed
  most significant first. ORACLE-TT might appear as 
  a literal symbol in the call -- this will be replaced with the
  oracle's truth table before execution. Alternatively, you might
  specify a fixed truth table as a space-separated, bracket-enclosed list specifying
  the right-hand side of the truth table.

- <b>(LIMITED-ORACLE ORACLE-TT \<max-calls> \<q1> \<q2> ... \<qn> \<out>)</b>:
  like ORACLE but with one additional argument, max-calls, which should
  be a positive integer. If the provided number of oracle calls has
  already been made by the time this instruction is executed then it
  will have no effect.

\subsection qgame_syntax_control Control Instructions

The following instructions do not specify gates, but rather influence the control
flow of the program or help with debugging.

- <b>(MEASURE \<q>) \<1-branch> (END) \<0-branch> (END)</b>:
  causes a measurement-based branch in the execution of the quantum
  program (causing the creation of a new QSys object). In one branch
  the state will be collapsed as if "1" was 
  read from the specified qubit, the 1-branch code will be executed
  and the 0-branch code will be skipped. In the other branch the
  state will be collapsed as if "0" was read from the specified qubit,
  the 0-branch code will be executed and the 1-branch code will be
  skipped. Measurement structures can be nested within the branches
  of other measurement structures. If there is no second END then
  execution in the 1 case terminates after execution of the 1-branch.
  If there is no first END then there is no 0-branch. Unmatched ENDs
  are ignored.
 
- <b>(HALT):</b> causes the executing quantum system to halt execution (to execute
  no further instructions)
  
- <b>(PRINTAMPS):</b> Prints the current state of the executing quantum system
  to STDOUT. This includes the measurement history, the current contents of the
  quantum register, and the probabilities for each amplitude.
  
- <b>(INSP):</b> This instruction only exists for compatibility with Lee Spector's
  LISP version of QGAME and does nothing in QGame++.
 */


