/***************************************************************************
 *   QGame++                                                               *
 *   =======                                                               *
 *   A Quantum Gate And Measurement Emulator.                              *
 *                                                                         *
 *   Copyright (C) 2003/04 by Manuel Nickschas                             *
 *   <qgame-devel@nickschas.de>                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef _QGAME___H_
#define _QGAME___H_
  
#include "qgtypes.h"
#include "qgerror.h"
#include "qggates.h"
#include "qgprog.h"

#include <string>
#include <vector>

namespace qgame {

  /** @defgroup toplevel Top Level Interfaces
   *  @{ */

  /** A result of a test run. This contains statistical data that has been produced
   *  during a test run (by calling QSys::testProgram()), i.e. after running the program
   *  for each specified test case and doing the final measurement after each run.
   */
  struct Result {
    int misses;           /**< Number of misses (probability of desired output below threshold) */
    double maxError;      /**< Maximum error that occured in any of the test cases              */
    double avgError;      /**< Average error occuring through all test cases                    */
    double maxExpOracles; /**< Maximum number of expected oracle calls for any test case        */
    double avgExpOracles; /**< Average number of expected oracle calls through all test cases   */
  };
  std::ostream & operator<<(std::ostream &, const Result &);
  
  /** This struct simply contains an outcome of a single measurement and is used to store the 
   *  measurement history of a system. The probability stored here is the probability with 
   *  which we could achieve this measurement result.
   */
  struct Measurement {
    int qubit;    /**< Qubit measured               */
    int result;   /**< Result measured              */
    double prob;  /**< Probability for this outcome */
  };

  /** A quantum system, the main interface to use QGame++. The core of a quantum system is a quantum
   *  register containing a certain number of qubits, which are being manipulated by a quantum
   *  program given as a QProgram object. A QSys object can be cloned (a functionality that 
   *  is mostly used for executing different branches after a measurement). There are two main
   *  ways to run a program; both involve calling a static method:
   *  - \em execute() performs a single run of the given quantum program, using the provided oracle
   *    truth table, and returns the list of resulting quantum systems (because each measurement
   *    causes a branch of the execution flow, and therefore produces an additional QSys object).
   *    Remember that you have to take care of deleting the QSys objects yourself if they are no
   *    no longer needed!
   *
   *  - \em testProgram() tests a program on a number of test cases, each consisting of an oracle
   *    truth table and the result one wants to receive by measurement of a number of qubits at
   *    the end of the run. The probability of this result is determined by measuring across each
   *    quantum that results from the execution of the program.
   *    A test run is considered successful if the probability of error is below a given threshold.
   *    This call returns those statistics as a Result.
   */
  class QSys {
    QProgram *prg;
    int instrPnt;
    double prob;
  
    int numQubits;
    Qureg reg;
    std::string oracleTT;
    int oracleCount;
    std::vector<Measurement> measureHistory;
    std::vector<bool> execInstr;  // for branch handling

    std::vector<QSys*> exec() throw(Error); // internal exec without init
    void markBranch(bool branch); // set execInstr and instrPnt for appropriate branch
    void appendMeasurement(QubitList q, int outcome, double prob); // append result of measurement to hist
    
    friend std::ostream& operator<<(std::ostream &, const QSys &);
    
   public:
    QSys();
    ~QSys();
    void init(int numQubits, QProgram *prg, const std::string &oracleTT);
            /**< Initialize a quantum system with the provided data, setting the register
             *   to |00..0> and installing the truth table to the oracle gates found in \em prg. */
    
    QSys *clone(double priorProb); /**< Return a new QSys object that is a clone of the current one,
                                    *   and has the given prior probability. */ 
    
    static std::vector<QSys*> execute(int numQubits, QProgram *prg, const std::string &oracleTT) throw(Error);
            /**< Execute the given program and return a list of pointers to the resulting QSys objects, each
             *   with its own measurement history and resulting prior probability.
             *   You have to take care of deleting the QSys objects yourself if they are no longer needed.
             */
    static Result testProgram(
                        int numQubits,                       /**< Number of qubits             */ 
                        QProgram *prg,                       /**< Pointer to program           */
                        const std::vector<TestCase> &cases,  /**< List of test cases           */
                        const QubitList &finalMeasure,       /**< Qubits for final measurement */
                        double threshold                     /**< Threshold for error          */
                  ) throw(Error); /**< Perform a number of test runs on the provided test cases as described
                                   *   in the class description, and return stats as a result. */
    
    Qureg  getReg();                  /**< Get contents of the system's quantum register */
    void   setReg(const Qureg &);     /**< Set the system's quantum register             */
    double getProb();                 /**< Get prior probability of quantum system       */
    void   setProb(double prob);      /**< Set prior probability of quantum system       */
    int    getOracleCount();          /**< Get the number of oracle calls in the history of this system. */
    std::vector<Measurement> getMeasurementHistory();   /**< Get the measurement history of the system.  */
    
    void apply(QGate *, const QubitList &) throw(Error); /**< Apply a QGate to the given qubits. */
    void apply(const QInstr &) throw(Error); /**< Apply an instruction given as a string. Instructions that do
                                              *   not contain gates, such as control flow instructions or
                                              *   measurements, will be ignored. */
  };
  
  std::ostream& operator<<(std::ostream &, const QSys &);
  /** @} */
};

#endif
