/***************************************************************************
 *   QGame++                                                               *
 *   =======                                                               *
 *   A Quantum Gate And Measurement Emulator.                              *
 *                                                                         *
 *   Copyright (C) 2003/04 by Manuel Nickschas                             *
 *   <qgame-devel@nickschas.de>                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _QGGATES_H_
#define _QGGATES_H_

#include "qgtypes.h"
#include "qgerror.h"

#include <vector>
#include <string>
#include <iostream>
#include <cmath>

namespace qgame {
  
  /** Base class for a quantum gate.
   *  All quantum gates are derived from this (virtual) base class and must implement
   *  numQb() and apply().
   */
  class QGate {

   public:
    virtual std::string dump() const;   /**< Return a string describing the gate */
    virtual unsigned int numQb() = 0;   /**< Number of qubits manipulated, NOT amplitudes */
    virtual void apply(QuSubReg &) = 0; /**< Apply gate to given qubits */
  };

  /** MATRIX-GATE #2A((row1)(row2)...(rown)): Defines a general matrix gate.
   *  This is also the base class for all other quantum gates that are defined
   *  by a (unitary) matrix.
   *  \ingroup (qgates)
   */
  class QMatrixGate : public QGate {
   protected:
    int numQubits;  /**< Number of affected Qubits */
    int numAmps;    /**< Number of amplitudes, generally 2^numQubits */
    std::vector<std::vector<Complex> > matrix;  /**< Unitary matrix [numAmps x numAmps] defining the gate */
    QMatrixGate() {} ;
   public:
    virtual unsigned int numQb();
    virtual void apply(QuSubReg &) throw(Error);
    
    std::string dump() const;
    QMatrixGate(std::vector<std::vector<Complex> >);  /**< Construct a matrix gate from a given matrix. */
  };
  
  std::ostream & operator<<(std::ostream &o, const QGate &);
  
  /****************************************************************************
   * Standard quantum gates
   ****************************************************************************/
  
  /** @defgroup qgates Quantum Gates
   *  @{ */ 
  
  /** QNOT \<q>: Applies a quantum NOT gate to the given qubit. */
  class QGateQNOT : public QMatrixGate {
   public:
    QGateQNOT();
    std::string dump() const;
  };
  
  /** CNOT \<control> \<target>: Applies a controlled NOT gate to the target qubit. */
  class QGateCNOT : public QMatrixGate {
   public:
    QGateCNOT();
    std::string dump() const;
  };
  
  /** NAND \<in1> \<in2> \<out>: Applies a quantum NAND gate to the specified qubits. */
  class QGateNAND : public QMatrixGate {
   public:
    QGateNAND();
    std::string dump() const;
  };
  
  /** SRN \<q>: Applies a quantum square-root-of-NOT gate to the given qubit. */
  class QGateSRN : public QMatrixGate {
   public:
    QGateSRN();
    std::string dump() const;
  };
  
  /** HADAMARD \<q>: Applies a Hadamard transformation to the given qubit. */
  class QGateH2 : public QMatrixGate {
   public:
    QGateH2();
    std::string dump() const;
  };
  
  /** U-THETA \<q> \<theta>: Rotates the given qubit by the (real-valued) angle \em theta. */
  class QGateUTHETA : public QMatrixGate {
    double theta;
   public:
    QGateUTHETA(double theta); /**< Construct gate with rotation angle \em theta. */
    std::string dump() const;
  };
  
  /** CPHASE-OLD \<control> \<target> \<alpha>: Old version of a controlled phase gate. */
  class QGateCPHASEOLD : public QMatrixGate {
    double alpha;
   public:
    QGateCPHASEOLD(double alpha); /**< Construct gate with a phase shift of \em alpha. */
    std::string dump() const;
  };
  
  /** CPHASE \<control> \<target> \<alpha>: Another, probably preferable version of a controlled
   *  phase gate. */
  class QGateCPHASE : public QMatrixGate {
   double alpha;
   public:
    QGateCPHASE(double alpha); /**< Construct gate with a phase shift of \em alpha. */
    std::string dump() const;
  };

  /** U2 \<q> \<phi> \<theta> \<psi> \<alpha>: A general rotation gate for a single qubit
   *  with four real-valued parameters. See the \ref qgame_lang_desc "QGame++ language description"
   *  for details. */
  class QGateU2 : public QMatrixGate {
    double phi,theta,psi,alpha;
   public:
    QGateU2(double phi, double theta, double psi, double alpha); /**< Construct U2 gate with given params. */
    std::string dump() const;
  };
  
  /** SWAP \<q1> \<q2>: Swaps the amplitudes of the specified qubits. */
  class QGateSWAP : public QMatrixGate {
   public:
    QGateSWAP();
    std::string dump() const;
  };
  
  /** ORACLE ORACLE-TT \<q1> \<q2> ... \<qn> \<out>: Applies a quantum oracle to the specified qubits. */
  class QOracle : public QMatrixGate {
    std::string truthtable;
   public:
    QOracle(const std::string &truthtable); /**< Construct a quantum oracle implementing the truth table
                                             *   given as a string of 0 and 1, specifying the outputs. */
    std::string dump() const;
  };

  /** @} */
};

#endif
