/***************************************************************************
 *   QGame++                                                               *
 *   =======                                                               *
 *   A Quantum Gate And Measurement Emulator.                              *
 *                                                                         *
 *   Copyright (C) 2003/04 by Manuel Nickschas                             *
 *   <qgame-devel@nickschas.de>                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _QGPROG_H_
#define _QGPROG_H_

#include "qgerror.h"
#include "qgtypes.h"
#include "qggates.h"
//#include "qgame++.h"

#include <string>
#include <vector>
#include <iostream>
#include <sstream>

/******************************************************************************
 * This file contains data related to quantum programs.
 ******************************************************************************/
namespace qgame {

  /** @defgroup qprogram Quantum Programs
   *  @{ */

  /** Instruction types. */
  enum { INVALID, COMMENT, NOOP, QNOT, CNOT, SRN, NAND, HADAMARD, U_THETA, CPHASE_OLD, CPHASE,
         U2, SWAP, ORACLE, LIMITED_ORACLE, MATRIX_GATE, MEASURE, END, HALT, PRINTAMPS, USER };
  
  /** This struct contains information about the parsed QProgram, such as the number of Qubits
   *  that should be used, or a list of test cases appropriate for the program.
   *  Such settings can be contained in the program file parsed by QProgram::read().
   */ 
  struct QProgramSettings {
    int numQb;                           /**< Number of Qubits the program uses              */
    std::vector<qgame::TestCase> cases;  /**< List of test cases appropriate for the program */
    qgame::QubitList finalQb;            /**< List of final measurement qubits               */
    double threshold;                    /**< Error probability threshold                    */
    QProgramSettings() { numQb = -1; threshold = -1; }
  };
         
  /** A struct resembling a single instruction of a QGame program.
   */  
  struct QInstr {
    int type;          /**< Instruction type */
    QubitList qubits;  /**< List of qubits the instruction affects */
    QGate *gate;       /**< The gate the instruction applies to the given qubits, if any; NULL otherwise */
    int extra;         /**< Used to store extra data, if needed. */
    bool flag;         /**< A flag that can be used if needed. */
    
    QInstr() { type = INVALID; gate = NULL; }
    ~QInstr() { delete gate; }
  };
  std::ostream & operator<<(std::ostream &, const QInstr &);

  /** A quantum program. Basically this stores a vector if QInstr objects and
   *  provides functions to read a QProgram from a string or a file.
   *  The interface functions as well as the syntax will most likely be changed in the
   *  future! */
  class QProgram {
    std::vector<QInstr*> instr;
    int instrPnt;
  
    QInstr *parseLine(const std::string &line) throw(Error);
    void parseSetting(QProgramSettings &set, const std::string &line) throw(Error);
    QubitList getQubits(int num, std::istringstream &) throw(Error);
    
    friend std::ostream & operator<<(std::ostream &, const QProgram &);
   public:
    QProgram();
    ~QProgram();
    
    unsigned int length() const;  /**< The length of the program (number of instructions) */
    
    QProgramSettings read(std::istream &stream)  throw(Error); /**< Read a program from a stream. Settings provided will be returned. 
                                                                    Note that the QProgramSettings object returned may contain empty settings! */
    bool addInstruction(const std::string &line) throw(Error); /**< Append an instruction given as a line of text. */
    bool addInstruction(QInstr*)                 throw(Error); /**< Append an instruction given as a QInstr object. */
    
    void installOracle(const std::string &oracleTT); /**< Install the given truth table to \em all oracle gates in the program. */
    
    QInstr * operator[](unsigned int) throw(Error);  /**< Return a reference to the addressed instruction. */
  };
  std::ostream & operator<<(std::ostream &, const QProgram &);
  
  /** @} */
};

#endif
