/***************************************************************************
 *   QGame++                                                               *
 *   =======                                                               *
 *   A Quantum Gate And Measurement Emulator.                              *
 *                                                                         *
 *   Copyright (C) 2003/04 by Manuel Nickschas                             *
 *   <qgame-devel@nickschas.de>                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef _QGTYPES_H_
#define _QGTYPES_H_

#include <iostream>
#include <vector>
#include <string>

#include "qgerror.h"

/***************************************************************************
 * Basic types for QGame++.
 ***************************************************************************/
 
namespace qgame {

  /** @defgroup basic_types Basic Types
   *  @{ */

  /** Class for complex numbers. This class represents complex numbers, mostly
   *  amplitudes, and provides basic functions and operators.
   *  This is, however, by no means a complete implementation for handling
   *  complex numbers, but instead what is needed for QGame++.
   */
  class Complex {
    double re,im;
    friend double sqabs(const Complex&);
    friend std::ostream& operator<<(std::ostream &, const Complex &);
 
   public:
    Complex(double re = 0, double im = 0);  /**< Constructor */
  
    double Re();        /**< Get real part of number.      */
    double Im();        /**< Get imaginary part of number. */
    double sqabs();     /**< Get squared absolute value.   */
  
    Complex operator+(const Complex&);  /**< Add two complex numbers. */
    Complex operator-(const Complex&);  /**< Subtract two complex numbers. */
    Complex operator*(const Complex&);  /**< Multiply two complex numbers. */
    Complex operator+=(const Complex&); /**< Add a complex number. */
    Complex operator*=(const Complex&); /**< Multiply with a complex number. */
  };
  std::ostream& operator<<(std::ostream &, const Complex &);
  double sqabs(const Complex&);   /**< Get squared absolute value of a complex number. */
  
  /** A list of qubits. This is used to store a list of qubits one wants to manipulate
   *  or otherwise access, as a vector of integers. For example, if you want to apply 
   *  an operator to qubits 3 and 0, first create a QubitList object,
   *  add 3 and 0, and then provide the list to the operator function. The least 
   *  significant (rightmost) qubit must be added first.
   */
  class QubitList {
    std::vector<int> qbl;
    friend std::ostream & operator<<(std::ostream &, const QubitList &);
    
   public:
    unsigned int size() const; /**< Get the size of the list (number of stored qubits).      */
    void append(int qbnum);    /**< Add one qubit to the list, LSB comes first.              */
    int operator[](const unsigned int) const throw(Error); 
                               /**< Get indexed list entry - this is read-only. */
  };
  std::ostream & operator<<(std::ostream &, const QubitList &);
  
  /** References one or more qubits of a Qureg.
   *  This enables access to only a part of a Qureg, addressed by one or more qubits.
   *  The entries of a QuSubReg (accessed by []) are in binary order, addressed by the
   *  qubits in the QubitList given to Qureg::getSubRegs(qubits), with the first qubit 
   *  added to the list being the least significant (rightmost, fastest varying...).
   *  Note that a QuSubReg contains \em references, thus manipulating an amplitude by means
   *  of QuSubReg[] will directly influence the underlying Qureg. Which is of course what
   *  one usually wants.
   */
  class QuSubReg {
    int num;
    std::vector<Complex*> sub;
    friend class Qureg;
    
   public:
    QuSubReg(int numQb); /**< The size of the QuSubReg (number of qubits, NOT amplitudes!) must be given. */
    ~QuSubReg();
    
    unsigned int numQb() const; /**< Returns the size of the QuSubReg (number of qubits, NOT amplitudes)  */
    unsigned int numAmps() const; /**< Returns the number of amplitudes */
    Complex& operator[](const unsigned int) throw(Error); /**< Returns a \em reference to the addressed qubit. */
  };
  
  /** A quantum register.
   *  This class represents a quantum register and provides manipulation functions. If you need
   *  to access a part of the Qureg as addressed by a number of qubits (in arbitrary order), use
   *  getSubRegs(QubitList). You can address individual amplitudes by using [].
   */
  class Qureg {
    int num;
    std::vector<Complex> amp;
    
    friend class QuSubReg;
    friend std::ostream& operator<<(std::ostream &, Qureg);
   public:
    Qureg(int size = 0); /**< The size of the Qureg (number of qubits, NOT amplitudes) must be given. */
   
    unsigned int size() const; /**< Returns the size of the Qureg (number of qubits, NOT amplitudes). */
    std::vector<QuSubReg> getSubRegs(const QubitList &qubits);
                          /**< Returns a vector of QuSubReg objects, that is, references to the amplitudes
                           *   referenced by the QubitList. Each QuSubReg object contains references to the
                           *   amplitudes of the qubits referenced by the given QubitList, with the first 
                           *   qubit number added to the list being the least signicant. The QuSubReg objects
                           *   are in binary order as addressed by the remaining qubits \em not in the QubitList,
                           *   with the rightmost of these qubits being the least significant.
                           *
                           *   For example, having a 3 qubit Qureg and a QubitList containing 1 (this means,
                           *   the middle qubit should be manipulated), you would get four QuSubReg objects:
                           *   { |000>, |010> } , { |001>, |011> }, { |100>, |110> }, { |101>, |111> }
                           */
    
    std::vector<double> getProbabilities(const QubitList &qubits); 
                          /**< Returns a list of probabilities for measuring the given qubits,
                           *   in binary order as given in the QubitList. */
    void collapse(const QubitList &qubits, int value);
                          /**< Collapses the given qubits to the value given (which can be thought as
                           *   a binary value corresponding to the given QubitList). No normalization
                           *   is done by now, although this might change in later versions of
                           *   QGame++ to provide a more physically accurate behavior.
                           */
    void normalize();     /**< Normalize the Qureg, so that all probabilities sum to 1. */
    Complex & operator[](const unsigned int) throw(Error); /**< Returns a reference to the addressed amplitude. */
  };
  std::ostream& operator<<(std::ostream &, Qureg);

  /** @} */
  
  /** \ingroup toplevel */
  /** A test case.
   *  This struct describes a single test case, consisting of (the right side of) an oracle truth table
   *  and the desired result one wants to measure through the final measurement qubits
   *  after running the program. The oracle truth table is given as a string of 0 and 1, 
   *  defining the output the oracle should produce for each input (sorted in increasing,
   *  binary order).
   */
  struct TestCase {
    std::string oracleTT; /**< Right-hand side of oracle truth table */
    int desiredRes;       /**< Desired result of final measurement   */
    TestCase(const std::string &tt, int res) { oracleTT = tt; desiredRes = res; }
                                       /**< Construct test case */
    TestCase(const std::string &casestr); /**< Construct from a param string: \<oraclett>-\<res>, e.g. 0100-1. 
                                           * desiredRes will be -1 if the string was invalid.
                                           */
                         
  };
};

#endif
